@@ -37,14 +37,16 @@ class Socket(object):
 
     def receive(self, pkt):
         """Receive packet from the client."""
-        self.server.logger.info('%s: Received packet %s with %s', self.sid,
-                                packet.packet_names[pkt.packet_type],
-                                pkt.data)
+        self.server.logger.info('%s: Received packet %s data %s',
+                                self.sid, packet.packet_names[pkt.packet_type],
+                                pkt.data if not isinstance(pkt.data, bytes)
+                                else '<binary>')
         if pkt.packet_type == packet.PING:
             self.last_ping = time.time()
             self.send(packet.Packet(packet.PONG, pkt.data))
         elif pkt.packet_type == packet.MESSAGE:
-            self.server._trigger_event('message', self.sid, pkt.data)
+            self.server._trigger_event('message', self.sid, pkt.data,
+                                       async=self.server.async_handlers)
         elif pkt.packet_type == packet.UPGRADE:
             self.send(packet.Packet(packet.NOOP))
         else:
@@ -54,15 +56,16 @@ class Socket(object):
         """Send a packet to the client."""
         if self.closed:
             raise IOError('Socket is closed')
-        if time.time() - self.last_ping > self.server.ping_interval * 5 / 4:
+        if time.time() - self.last_ping > self.server.ping_timeout:
             self.server.logger.info('%s: Client is gone, closing socket',
                                     self.sid)
             self.close(wait=False, abort=True)
             return
         self.queue.put(pkt)
-        self.server.logger.info('%s: Sending packet %s with %s', self.sid,
-                                packet.packet_names[pkt.packet_type],
-                                pkt.data)
+        self.server.logger.info('%s: Sending packet %s data %s',
+                                self.sid, packet.packet_names[pkt.packet_type],
+                                pkt.data if not isinstance(pkt.data, bytes)
+                                else '<binary>')
 
     def handle_get_request(self, environ, start_response):
         """Handle a long-polling GET request from the client."""
@@ -95,12 +98,13 @@ class Socket(object):
 
     def close(self, wait=True, abort=False):
         """Close the socket connection."""
-        self.server._trigger_event('disconnect', self.sid)
-        if not abort:
-            self.send(packet.Packet(packet.CLOSE))
-        self.closed = True
-        if wait:
-            self.queue.join()
+        if not self.closed:
+            self.server._trigger_event('disconnect', self.sid, async=False)
+            if not abort:
+                self.send(packet.Packet(packet.CLOSE))
+            self.closed = True
+            if wait:
+                self.queue.join()
 
     def _upgrade_websocket(self, environ, start_response):
         """Upgrade the connection from polling to websocket."""
@@ -127,7 +131,7 @@ class Socket(object):
                                         always_bytes=False):
                 self.server.logger.info(
                     '%s: Failed websocket upgrade, no PING packet', self.sid)
-                return
+                return []
             ws.send(packet.Packet(
                 packet.PONG,
                 data=six.text_type('probe')).encode(always_bytes=False))
@@ -141,24 +145,27 @@ class Socket(object):
                     ('%s: Failed websocket upgrade, expected UPGRADE packet, '
                      'received %s instead.'),
                     self.sid, pkt)
-                return
+                return []
             self.upgraded = True
         else:
             self.connected = True
             self.upgraded = True
 
+        # start separate writer thread
         def writer():
             while True:
                 try:
                     packets = self.poll()
                 except IOError:
                     break
+                if not packets:
+                    # empty packet list returned -> connection closed
+                    break
                 try:
                     for pkt in packets:
                         ws.send(pkt.encode(always_bytes=False))
                 except:
                     break
-
         self.server.start_background_task(writer)
 
         self.server.logger.info(
@@ -170,6 +177,7 @@ class Socket(object):
             except:
                 break
             if p is None:
+                # connection closed by client
                 break
             if isinstance(p, six.text_type):  # pragma: no cover
                 p = p.encode('utf-8')
@@ -178,5 +186,8 @@ class Socket(object):
                 self.receive(pkt)
             except ValueError:
                 pass
+
         self.close(wait=True, abort=True)
         self.queue.put(None)  # unlock the writer task so that it can exit
+
+        return []
